import 'package:grid_world/grid_world.dart';
import 'package:test/test.dart';

void main() {
  final halfArrow = GridWorld.fromString('''
..#..
..##.
..###
..#..
..#..
..#..
''');

  final blinker = GridWorld.fromString('''
.....
..#..
..#..
..#..
.....
''');

  test('construction from string', () {
    expect(halfArrow.toString(), equals('''
..#..
..##.
..###
..#..
..#..
..#..
'''));
  });

  final _identity06 = GridWorld.identity(6);
  final _identity12 = GridWorld.identity(12);

  test('equals', () {
    final other = GridWorld.identity(6);
    expect(_identity06, isNot(same(other)));
    expect(_identity06, equals(other));
    expect(_identity06, isNot(equals(other.clockwise90())));
  });

  test('hashCode', () {
    final other = GridWorld.identity(6);
    expect(_identity06.hashCode, equals(other.hashCode));
    expect(_identity06, isNot(equals(other.clockwise90().hashCode)));
  });

  test('identity', () {
    expect(_identity06.toString(), equals('''
#.....
.#....
..#...
...#..
....#.
.....#
'''));
  });

  test('paste', () {
    var w = _identity12.paste(0, 0, halfArrow);
    expect(w.toString(), equals('''
..#.........
..##........
..###.......
..#.........
..#.........
..#..#......
......#.....
.......#....
........#...
.........#..
..........#.
...........#
'''));
    w = _identity12.paste(0, 3, halfArrow);
    expect(w.toString(), equals('''
#....#......
.#...##.....
..#..###....
.....#......
.....#......
.....#......
......#.....
.......#....
........#...
.........#..
..........#.
...........#
'''));
    w = _identity12.paste(4, 4, halfArrow);
    expect(w.toString(), equals('''
#...........
.#..........
..#.........
...#........
......#.....
......##....
......###...
......#.....
......#.....
......#..#..
..........#.
...........#
'''));
    w = _identity12.paste(12, 12, halfArrow);
    expect(w.toString(), equals('''
#................
.#...............
..#..............
...#.............
....#............
.....#...........
......#..........
.......#.........
........#........
.........#.......
..........#......
...........#.....
..............#..
..............##.
..............###
..............#..
..............#..
..............#..
'''));
  });

  test('counterClockwise90', () {
    var w = halfArrow.counterClockwise90();
    expect(w.toString(), equals('''
..#...
.##...
######
......
......
'''));
    expect(
        w
            .counterClockwise90()
            .counterClockwise90()
            .counterClockwise90()
            .toString(),
        equals(halfArrow.toString()));
  });

  test('clockwise90', () {
    var w = halfArrow.clockwise90();
    expect(w.toString(), equals('''
......
......
######
...##.
...#..
'''));
    expect(w.clockwise90().clockwise90().clockwise90().toString(),
        equals(halfArrow.toString()));
  });

  test('transpose', () {
    var w = halfArrow.transpose();
    expect(w.toString(), equals('''
......
......
######
.##...
..#...
'''));
    expect(_identity12.transpose().toString(), equals(_identity12.toString()));
    w = GridWorld.fromString('''
...#...
..#.#..
.#.#.#.
...#...
...#...
...#...
''').transpose();
    expect(w.toString(), equals('''
......
..#...
.#....
#.####
.#....
..#...
......
'''));
  });

  test('copy', () {
    var w = halfArrow.copy();
    expect(w, equals(halfArrow));
  });

  test('leftPadded', () {
    var w = halfArrow.padLeft(2);
    expect(w.toString(), equals('''
....#..
....##.
....###
....#..
....#..
....#..
'''));
  });

  test('rightPadded', () {
    var w = halfArrow.padRight(2);
    expect(w.toString(), equals('''
..#....
..##...
..###..
..#....
..#....
..#....
'''));
  });

  test('lrPadded', () {
    var w = halfArrow.lrPadded(2);
    expect(w.toString(), equals('''
....#....
....##...
....###..
....#....
....#....
....#....
'''));
  });

  test('topPadded', () {
    var w = halfArrow.padTop(2);
    expect(w.toString(), equals('''
.....
.....
..#..
..##.
..###
..#..
..#..
..#..
'''));
  });

  test('bottomPadded', () {
    var w = halfArrow.padBottom(2);
    expect(w.toString(), equals('''
..#..
..##.
..###
..#..
..#..
..#..
.....
.....
'''));
  });

  test('tbPadded', () {
    var w = halfArrow.tbPadded(2);
    expect(w.toString(), equals('''
.....
.....
..#..
..##.
..###
..#..
..#..
..#..
.....
.....
'''));
  });

  test('padded', () {
    var w = halfArrow.padded(3);
    expect(w.toString(), equals('''
...........
...........
...........
.....#.....
.....##....
.....###...
.....#.....
.....#.....
.....#.....
...........
...........
...........
'''));
  });

  test('expandToFit0', () {
    var w = blinker.expandToFit(blinker.nCols, blinker.nRows);
    expect(w, equals(blinker));
  });

  test('expandToFit1', () {
    expect(
          () => blinker.expandToFit(blinker.nCols, blinker.nRows-1),
      throwsA(
        isA<ArgumentError>().having(
              (error) => error.message,
          'should throw if world too tall',
          'world too tall (5) to expand into 4',
        ),
      ),
    );
    expect(
          () => blinker.expandToFit(blinker.nCols-1, blinker.nRows),
      throwsA(
        isA<ArgumentError>().having(
              (error) => error.message,
          'should throw if world too wide',
          'world too wide (5) to expand into 4',
        ),
      ),
    );
  });

  test('expandToFit2', () {
    var w = blinker.expandToFit(8, 13);
    expect(w.toString(), equals('''
........
........
........
........
........
...#....
...#....
...#....
........
........
........
........
........
'''));
  });

  test('expandToFit3', () {
    var w = blinker.expandToFit(5, 6);
    expect(w.toString(), equals('''
.....
..#..
..#..
..#..
.....
.....
'''));
  });

  test('append', () {
    var w = halfArrow;
    w = w.appendRight(w).appendRight(w).appendRight(w);
    expect(w.toString(), equals('''
..#....#....#....#..
..##...##...##...##.
..###..###..###..###
..#....#....#....#..
..#....#....#....#..
..#....#....#....#..
'''));
    expect(w.appendBottom(w).toString(), equals('''
..#....#....#....#..
..##...##...##...##.
..###..###..###..###
..#....#....#....#..
..#....#....#....#..
..#....#....#....#..
..#....#....#....#..
..##...##...##...##.
..###..###..###..###
..#....#....#....#..
..#....#....#....#..
..#....#....#....#..
'''));
  });

  test('irregularAppendRight', () {
    var w = halfArrow;
    w = w.appendRight(w.padTop(3));
    expect(w.toString(), equals('''
..#.......
..##......
..###.....
..#....#..
..#....##.
..#....###
.......#..
.......#..
.......#..
'''));
  });

  test('irregularAppendBottom', () {
    var w = halfArrow;
    w = w.appendRight(w).appendRight(w);
    w = w.appendBottom(w.padLeft(3));
    expect(w.toString(), equals('''
..#....#....#.....
..##...##...##....
..###..###..###...
..#....#....#.....
..#....#....#.....
..#....#....#.....
.....#....#....#..
.....##...##...##.
.....###..###..###
.....#....#....#..
.....#....#....#..
.....#....#....#..
'''));
  });

  test('bigX', () {
    var v = _identity06.appendRight(_identity06.clockwise90());
    var x = v.appendBottom(v.clockwise90().clockwise90());
    expect(x.toString(), equals('''
#..........#
.#........#.
..#......#..
...#....#...
....#..#....
.....##.....
.....##.....
....#..#....
...#....#...
..#......#..
.#........#.
#..........#
'''));
  });

  test('mixItUp1', () {
    var r1 = halfArrow.padLeft(2).padTop(2).padBottom(3).padRight(10);
    var r2 = r1.clockwise90().clockwise90();
    var w = r1.appendBottom(r2);
    expect(w.toString(), equals('''
.................
.................
....#............
....##...........
....###..........
....#............
....#............
....#............
.................
.................
.................
.................
.................
.................
............#....
............#....
............#....
..........###....
...........##....
............#....
.................
.................
'''));
    expect(w.transpose().toString(), equals('''
......................
......................
......................
......................
..######..............
...##.................
....#.................
......................
......................
......................
.................#....
.................##...
..............######..
......................
......................
......................
......................
'''));
  });

  test('mixItUp2', () {
    var w = halfArrow.padRight(2);
    var w1 = w.appendRight(w.clockwise90());
    var w2 = w.counterClockwise90().appendRight(w);
    w = w1.appendBottom(w2);
    expect(w.toString(), equals('''
..#..........
..##.........
..###..######
..#.......##.
..#.......#..
..#..........
.............
........#....
........##...
..#.....###..
.##.....#....
######..#....
........#....
.............
'''));
  });

  /// Check a three step iterable.
  test('gridworlditerable1', () {
    var itr = GridWorldIterable(blinker, limit: 3).iterator;
    assert(itr.moveNext());
    expect(itr.current.toString(), equals('''
.....
..#..
..#..
..#..
.....
'''));
    assert(itr.moveNext());
    expect(itr.current.toString(), equals('''
.....
.....
.###.
.....
.....
'''));
    assert(itr.moveNext());
    expect(itr.current.toString(), equals('''
.....
..#..
..#..
..#..
.....
'''));
    assert(!itr.moveNext());
  });

  /// Another iterable check, this time with snapshots to make a list.
  test('gridworlditerable2', () {
    List<GridWorld> list = List<GridWorld>.from(
        GridWorldIterable(blinker, limit: 3, snapshots: true));
    expect(list.length, equals(3));
    expect(list[0].toString(), equals('''
.....
..#..
..#..
..#..
.....
'''));
    expect(list[1].toString(), equals('''
.....
.....
.###.
.....
.....
'''));
    expect(list[2].toString(), equals('''
.....
..#..
..#..
..#..
.....
'''));
  });
}
